from py2neo.client import ConnectionUnavailable

from smdc.format import Format
from smdc.note import Note
import io
from smdc.format.cypher import escape_cypher
from smdc.parse import obsidian_url, PROP_VAULT, PROP_PATH
from py2neo import Graph, Node, Relationship, Subgraph
from py2neo.database.work import ClientError
from pathlib import Path
import tqdm

CAT_DANGLING = "SMD_dangling"
CAT_NO_TAGS = "SMD_no_tags"

PROP_COMMUNITY = "SMD_community"
INDEX_PROPS = ['name', 'aliases']

def get_community(note: Note, communities: [str], community_type: str):
    if community_type == "tags":
        if note.tags:
            community = escape_cypher(note.tags[0])
        else:
            community = CAT_NO_TAGS
    elif community_type == "folders":
        community = str(Path(note.properties[PROP_PATH]).parent)
        if community not in communities:
            communities.append(community)
    return communities.index(community)

def node_from_note(note: Note, all_tags: [str], all_communities: [str], community_type: str) -> Node:
    tags = [CAT_NO_TAGS]
    if note.tags:
        tags = list(map(escape_cypher, note.tags))
        for tag in tags:
            if tag not in all_tags:
                all_tags.append(tag)
    properties = {}
    for property, value in note.properties.items():
        properties[property] = escape_cypher(str(value))
    properties[PROP_COMMUNITY] = get_community(note, all_communities, community_type)
    return Node(*tags, **properties)

def add_rels_between_nodes(rels, src_node, trgt_node, subgraph: [Relationship]):
    # Adds all relations between src node and trgt node as described in rels to subgraph
    for rel in rels:
        properties = {}
        for property, value in rel.properties.items():
            properties[property] = escape_cypher(str(value))
        subgraph.append(Relationship(src_node, escape_cypher(rel.type), trgt_node, **properties))


def create_index(graph, tag):
    try:
        for prop in INDEX_PROPS:
            graph.run(f"CREATE INDEX index_{prop}_{tag} IF NOT EXISTS FOR (n:{tag}) ON (n.{prop})")
        graph.run(f"CREATE INDEX index_name_vault IF NOT EXISTS for (n:{tag}) ON (n.{prop})")
    except ClientError as e:
        print(e)
        print(f"Warning: Could not create index for {tag}", flush=True)

def create_dangling(name:str, vault_name:str, all_communities: [str]) -> Node:
    n = Node(CAT_DANGLING, name=escape_cypher(name), community=all_communities.index(CAT_DANGLING),
                                 obsidian_url=escape_cypher(obsidian_url(name, vault_name)))
    n[PROP_VAULT] = vault_name
    return n

class Neo4j(Format):

    def parse(self, file: io.TextIOWrapper, name, parsed_notes: [Note], args) -> Note:
        raise NotImplementedError

    def write(self, parsed_notes: [Note], args):
        try:
            g = Graph(password=args.password)
        except ClientError as e:
            print("invalid user credentials", flush=True)
            raise e
        except ConnectionUnavailable as e:
            print("no connection to db", flush=True)
            raise e
        tx = g.begin()
        if not args.retaindb:
            print("Clearing neo4j database")
            tx.run(f"MATCH (n) WHERE n.{PROP_VAULT}='{args.vault_name}' DETACH DELETE n")

        nodes = {}
        print("Converting nodes", flush=True)
        all_tags = [CAT_DANGLING, CAT_NO_TAGS]
        all_communities = all_tags if args.community == "tags" else [CAT_DANGLING]
        # First create all nodes in the graph before doing the relationships, so they all exist.
        for name, note in tqdm.tqdm(parsed_notes.items()):
            node = node_from_note(note, all_tags, all_communities, args.community)
            nodes[name] = node

        if nodes:
            print("Transferring nodes to graph", flush=True)
            tx.create(Subgraph(nodes=nodes.values()))

        rels_to_create = []
        nodes_to_create = []
        print("Creating relationships", flush=True)
        i = 1
        for name, note in tqdm.tqdm(parsed_notes.items()):
            if not note.out_rels.keys():
                continue
            src_node = nodes[name]
            for trgt, rels in note.out_rels.items():
                if trgt not in nodes:
                    nodes[trgt] = create_dangling(trgt, args.vault_name, all_communities)
                    nodes_to_create.append(nodes[trgt])
                trgt_node = nodes[trgt]
                add_rels_between_nodes(rels, src_node, trgt_node, rels_to_create)
            # Send batches to server. Greatly speeds up conversion.
            if i % args.batch_size == 0:
                tx.create(Subgraph(nodes=nodes_to_create, relationships=rels_to_create))
                rels_to_create = []
                nodes_to_create = []
            i += 1
        if rels_to_create or nodes_to_create:
            tx.create(Subgraph(nodes=nodes_to_create, relationships=rels_to_create))
        print("Committing data", flush=True)
        tx.commit()

        print("Creating index", flush=True)
        # TODO: Schema inference for auto-indexing?
        for tag in tqdm.tqdm(all_tags):
            create_index(g, tag)
        try:
            g.run("CALL db.index.fulltext.drop(\"SMDnameAlias\")")
        except ClientError:
            pass
        try:
            g.run("CALL db.index.fulltext.drop(\"SMDcontent\")")
        except ClientError:
            pass
        if all_tags:
            g.run("CALL db.index.fulltext.createNodeIndex(\"SMDnameAlias\", [\"" + "\", \"".join(all_tags) + "\"], [\"name\", \"aliases\"])")
            if args.index_content:
                g.run("CALL db.index.fulltext.createNodeIndex(\"SMDcontent\", [\"" + "\", \"".join(all_tags) + "\"], [\"content\"])")
        return g, all_tags, all_communities





import yaml
import os
import markdown
from markdown.extensions import Extension
from markdown.preprocessors import Preprocessor
import re

def parse_yaml_header(file):
    lines = []
    line = file.readline()
    while line != "---" + os.linesep and line:
        lines.append(line)
        line = file.readline()

    return yaml.safe_load("".join(lines))


class HashtagExtension(Extension):
    # Code based on https://github.com/Kongaloosh/python-markdown-hashtag-extension/blob/master/markdown_hashtags/markdown_hashtag_extension.py
    # Used to extract tags from markdown
    def extendMarkdown(self, md):
        """ Add FencedBlockPreprocessor to the Markdown instance. """
        md.registerExtension(self)
        md.preprocessors.register(HashtagPreprocessor(md), 'hashtag', 10) # After HTML Pre Processor


class HashtagPreprocessor(Preprocessor):
    ALBUM_GROUP_RE = re.compile(
            r"""(?:(?<=\s)|^)#(\w*[A-Za-z_]+\w*)"""
    )

    def __init__(self, md):
        super(HashtagPreprocessor, self).__init__(md)

    def run(self, lines):
        """ Match and store Fenced Code Blocks in the HtmlStash. """
        HASHTAG_WRAP = '''<a href="#{0}" class="tag"> #{0}</a>'''
        text = "\n".join(lines)
        while True:
            hashtag = ''
            m = self.ALBUM_GROUP_RE.search(text)
            if m:
                hashtag += HASHTAG_WRAP.format(m.group()[1:])
                placeholder = self.markdown.htmlStash.store(hashtag)
                text = '%s %s %s' % (text[:m.start()], placeholder, text[m.end():])
            else:
                break
        return text.split('\n')


# def makeExtension(*args, **kwargs):
#     return HashtagExtension(*args, **kwargs)

def markdownToHtml(md_text):
    return markdown.markdown(md_text, extensions=['mdx_wikilink_plus','fenced_code',
                                                  'footnotes', 'tables', HashtagExtension()],
                             extension_configs={
                                 'mdx_wikilink_plus': {
                                     'html_class': 'internal-link',
                                     'url_whitespace': ' '
                                 }
                             })

PUNCTUATION = ["#", "$", "!", ".", ",", "?", "/", ":", ";", "`", " ", "-", "+", "=", "|", os.linesep] + [str(i) for i in range(0, 10)]


def get_tags_from_line(line) -> [str]:
    pos_tags = [i for i, char in enumerate(line) if char == '#']
    tags = []
    for i in pos_tags:
        if i == 0 or line[i - 1] == ' ':
            index = next((index for index, c in enumerate(line[i+1:]) if c in PUNCTUATION), -1)
            if index == -1:
                tags.append(line[i+1:])
            else:
                tag = line[i + 1:index + i + 1]
                if len(tag) > 0:
                    tags.append(tag)
    return tags

def parse_wikilink(between_brackets:str, note_title: str) -> str:
    first_arg = between_brackets.split("|")[0]
    if len(first_arg) != 0:
        title = first_arg.split("#")[0]
        if len(title) == 0:
            # Wikilinks like [[#header]] refer to itself
            return note_title
        else:
            # Wikilinks like [[title#header]]
            return title
    return ""

def get_wikilinks_from_line(line, note_title) -> [str]:
    result = re.findall('\[\[(.*?)\]\]', line)
    if result:
        r = []
        for wikilink in result:
            title = parse_wikilink(wikilink, note_title)
            if title:
                r.append(title)
        return r
    return []

def escape_quotes(string) -> str:
    r1 = string.replace("\\\'", "\\\\\'")
    r1 = r1.replace("'", "\\\'")
    return r1.replace('\"', "\\\"")



from smdc import server_args, convert, parse_note, FORMAT_TYPES, Note, note_name, obsidian_url
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer
import time
from py2neo import Node, Subgraph, Relationship
from py2neo.data import walk
from smdc.format.neo4j import node_from_note, add_rels_between_nodes, CAT_DANGLING, CAT_NO_TAGS, create_index, \
    create_dangling, PROP_COMMUNITY, get_community
from smdc.format.cypher import escape_cypher
from pathlib import Path
import smdc
from smdc.parse import PROP_VAULT, PROP_PATH


def wrapper(fn):
    def _return(event):
        try:
            fn(event)
        except BaseException as e:
            print(e)
    return _return


class SMDSEventHandler():
    def __init__(self, graph, tags: [str], communities: [str], args):
        self.graph = graph
        self.nodes = graph.nodes
        self.relationships = graph.relationships
        self.args = args
        self.input_format = FORMAT_TYPES[args.input_format]
        self.vault_name = args.vault_name
        self.index_content = args.index_content
        self.tags = tags
        self.communities = communities

    def _clear_outgoing(self, node: Node):
        rels = self.relationships.match([node, None])
        if len(rels) > 0:
            self.graph.separate(Subgraph(relationships=rels))

    def _process_node_on_graph(self, note: Note):
        if smdc.DEBUG:
            print(note, flush=True)
        in_graph = self.nodes.match(**{'name': note.name, PROP_VAULT: self.vault_name})
        if len(in_graph) == 0:
            # Create new node
            node = node_from_note(note, self.tags, self.communities, self.args.community)
            if smdc.DEBUG:
                print("creating")
                print(node, flush=True)
            self.graph.create(node)
            return
        # Update
        node = in_graph.first()
        if smdc.DEBUG:
            print("updating")
            print(node, flush=True)
        # Update labels
        node.clear_labels()
        note_tags = [CAT_NO_TAGS]
        if note.tags:
            note_tags = list(map(escape_cypher, note.tags))
        node.update_labels(note_tags)
        for tag in note_tags:
            if tag not in self.tags:
                create_index(self.graph, tag)
                self.tags.append(tag)
        # Update properties
        node.clear()
        escaped_properties = {}
        for key, value in note.properties.items():
            escaped_properties[key] = escape_cypher(str(value))
        escaped_properties[PROP_COMMUNITY] = get_community(note, self.communities, self.args.community)
        node.update(escaped_properties)
        self.graph.push(node)

        # # Delete active relations
        # self._clear_outgoing(node)

        # Insert up-to-date relations
        rels_to_create = []
        nodes_to_create = []
        active_rels = list(self.relationships.match([node, None]))
        for trgt, rels in note.out_rels.items():
            trgt_node = self.nodes.match(**{'name': trgt, PROP_VAULT: self.vault_name})
            if len(trgt_node) == 0:
                trgt_node = create_dangling(trgt, self.vault_name, self.tags)
                nodes_to_create.append(trgt_node)
            else:
                trgt_node = trgt_node.first()
            # Possibly refactor this with
            for i, rel in enumerate(rels):
                properties = {}
                for property, value in rel.properties.items():
                    properties[property] = escape_cypher(str(value))
                rel_type = escape_cypher(rel.type)
                rem_rel = None
                compare_rel = Relationship(node, rel_type, trgt_node)
                # Update instead of removing makes sure the relationship has a persistent id
                for active_rel in active_rels:
                    walks = list(walk(active_rel))
                    if type(active_rel).__name__ == rel_type and walks[2] == trgt_node:
                    # if compare_rel == active_rel:
                        # Maybe this can leave dangling properties? But that'' an edge case. Not sure how to clear properties.
                        active_rel.clear()
                        active_rel.update(properties)
                        self.graph.push(active_rel)
                        rem_rel = active_rel
                        break
                if rem_rel is not None:
                    active_rels.remove(rem_rel)
                else:
                    rels_to_create.append(Relationship(node, rel_type, trgt_node, **properties))
        if rels_to_create or nodes_to_create:
            self.graph.create(Subgraph(nodes=nodes_to_create, relationships=rels_to_create))
        if len(active_rels) > 0:
            ids = []
            for active_rel in active_rels:
                ids.append(active_rel.identity)
            self.graph.separate(Subgraph(relationships=active_rels))
            print("onSMDRelDeletedEvent/" + "/".join(map(str, ids)))
            # for active_rel in active_rels:
            #     print(dir(active_rel))
            #     print(active_rel)
            #     print(f"onSMDRelDeletedEvent/{active_rel.identity}")

    def on_created(self):
        def _on_created(event):
            if smdc.DEBUG:
                print("On created", event.src_path, flush=True)
            # TODO: What if this name already exists in the vault? Does it make sense to override old data?
            note = parse_note(self.input_format, event.src_path, self.args)
            self._process_node_on_graph(note)
        return wrapper(_on_created)

    def on_deleted(self):
        def _on_deleted(event):
            if smdc.DEBUG:
                print("On deleted", event.src_path, flush=True)
            name = note_name(event.src_path)
            node = self.nodes.match(name=name).first()
            node_id = node.identity
            in_rels = self.relationships.match([None, node])
            if len(in_rels) > 0:
                # If there are still active incoming links, keep the node as a reference
                node.clear()
                node.clear_labels()
                node.add_label(CAT_DANGLING)
                node.name = escape_cypher(name)
                node.obsidian_url = escape_cypher(obsidian_url(name, self.vault_name))
                self._clear_outgoing(node)
            else:
                self.graph.delete(node)
            print(f"onSMDDeletedEvent/{node_id}", flush=True)
        return wrapper(_on_deleted)

    def on_modified(self):
        def _on_modified(event):
            if smdc.DEBUG:
                print("On modified", event.src_path, flush=True)
            note = parse_note(self.input_format, event.src_path, self.args)
            self._process_node_on_graph(note)
            print(f"onSMDModifyEvent/{note.name}", flush=True)
        return wrapper(_on_modified)

    def on_moved(self):
        def _on_moved(event):
            if smdc.DEBUG:
                print("On moved", event.src_path, event.dest_path, flush=True)
            old_name = note_name(event.src_path)
            node = self.nodes.match(name=old_name).first()
            new_name = note_name(event.dest_path)
            # TODO: What if this name already exists in the vault?
            node['name'] = new_name
            node['obsidian_url'] = obsidian_url(new_name, self.vault_name)
            node[PROP_PATH] = event.dest_path
            self.graph.push(node)
            print(f"onSMDMovedEvent/{old_name}/{new_name}", flush=True)
        return wrapper(_on_moved)

def stream(graph, tags, communities, args):
    # Code credit: http://thepythoncorner.com/dev/how-to-create-a-watchdog-in-python-to-look-for-filesystem-changes/
    event_handler = PatternMatchingEventHandler(patterns=["*.md"], case_sensitive=True)

    smds_event_handler = SMDSEventHandler(graph, tags, communities, args)
    event_handler.on_created = smds_event_handler.on_created()
    event_handler.on_deleted = smds_event_handler.on_deleted()
    event_handler.on_modified = smds_event_handler.on_modified()
    event_handler.on_moved = smds_event_handler.on_moved()

    observer = Observer()
    path = Path(args.input)
    if smdc.DEBUG:
        print(path.absolute(), flush=True)
    observer.schedule(event_handler, path=Path(args.input), recursive=args.r)

    observer.start()

    try:
        print("Stream is active!", flush=True)
        import sys
        sys.stdout.flush()
        while True:
            # TODO: Catch when connection to neo4j server is down.
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
        observer.join()

def main():
    args = server_args()
    args.output_format = 'neo4j'
    # Initialize the database
    graph, tags, communities = convert(args)
    # return
    # Start the server
    stream(graph, tags, communities, args)


if __name__ == "__main__":
    main()



from pathlib import Path
from smdc.format import Format
from tqdm import tqdm
import os
from urllib.parse import quote

PROP_OBSIDIAN_URL = "obsidian_url"
PROP_PATH = "SMD_path"
PROP_VAULT = "SMD_vault"

def note_name(path, extension=".md"):
    return os.path.basename(path)[:-len(extension)]

def obsidian_url(name:str, vault:str) -> str:
    return "obsidian://open?vault=" + quote(vault) + "&file=" + quote(name) + ".md"

def parse_note(format: Format, note_path, args):
    name = note_name(note_path)
    with open(Path(note_path), 'r', encoding='utf-8') as f:
        # TODO: This isn't passing parsed notes right now. But this isn't currently used.
        note = format.parse(f, name, [], args)
        # Assign automatic properties for handling data and plugins
        note.properties[PROP_OBSIDIAN_URL] = obsidian_url(name, args.vault_name)
        note.properties[PROP_PATH] = note_path
        note.properties[PROP_VAULT] = args.vault_name
        return note

def parse_folder(format: Format, args):
    notes_path = args.input
    note_extension = args.extension
    vault_name = args.vault_name
    if args.r:
        iterate = Path(notes_path).rglob("*" + note_extension)
    else:
        iterate = Path(notes_path).glob("*" + note_extension)
    all_files = list(iterate)
    parsed_notes = {}
    print("Parsing notes", flush=True)
    for path in tqdm(all_files):
        with open(path, mode='r', encoding='utf-8') as f:
            name = note_name(path, note_extension)
            try:
                note = format.parse(f, name, parsed_notes, args)
                note.properties[PROP_OBSIDIAN_URL] = obsidian_url(name, vault_name)
                note.properties[PROP_PATH] = path
                note.properties[PROP_VAULT] = vault_name
                parsed_notes[name] = note
            except Exception as e:
                print(e)
                print("Exception raised during parsing " + str(path) + ". Skipping this note! Please report this.", flush=True)
    print("Finished parsing notes", flush=True)
    return parsed_notes




DEBUG = False
from .note import Note, Relationship
from .args import convert_args, server_args
from smdc.format import FORMAT_TYPES
from .convert import convert
from .parse import parse_folder, parse_note, note_name, obsidian_url
from .stream import stream, main




import argparse
from pathlib import Path

import smdc

def _mutual_args(parser):
    parser.add_argument('--input', metavar='i', type=str, default=".",
                        help='directory with markdown files')
    parser.add_argument('--input_format', metavar='f', type=str, default="typed_list",
                        help='format of inputs')
    parser.add_argument('--extension', metavar='e', type=str, default=".md",
                        help='extension of markdown files')
    parser.add_argument('--retaindb', action="store_true", default=False)
    parser.add_argument('--password', metavar='p', type=str, default="")
    parser.add_argument("--vault_name", metavar='n', type=str, default=None, help="Defaults to input directory name")
    parser.add_argument("--batch_size", metavar='b', type=int, default=75, help="Batch size for sending to neo4j")
    parser.add_argument("--debug", action="store_true", help="Debug mode")
    parser.add_argument("--index_content", action="store_true", help="Use to index the content in neo4j. Can highly impact performance")
    parser.add_argument("--typed_links_prefix", metavar='t', type=str, default="-")
    parser.add_argument('--community', metavar='c', type=str, default="tags", help="Options: {tags, folders, none}. Used for color coding in Obsidian plugin.")
    parser.add_argument('--convert_markdown', action="store_true", help="Converts content property to HTML. Takes quite a bit longer on startup. ")
    parser.add_argument('-r', action="store_false", default=True)

def server_args():
    parser = argparse.ArgumentParser(
        description='Stream changes in Obsidian vaults to active neo4j server.')
    _mutual_args(parser)
    args = parser.parse_args()
    if not args.vault_name:
        args.vault_name = Path(args.input).name
    if args.debug:
        smdc.DEBUG = True
        print(args)
    return args

def convert_args():
    parser = argparse.ArgumentParser(
        description='Convert different typed links representations of Markdown into other formats.')
    _mutual_args(parser)
    parser.add_argument('--output_format', metavar='r', type=str, default="neo4j",
                        help='format of inputs')
    parser.add_argument('--output', metavar='o', type=str, default="out",
                        help='directory for output files')
    args = parser.parse_args()
    if not args.vault_name:
        args.vault_name = Path(args.input).name
    if args.debug:
        smdc.DEBUG = True
        print(args)
    return args
